"""Top-level package for bidsMReye."""
from __future__ import annotations

__author__ = """Remi Gau"""
__email__ = "remi.gau@gmail.com"

from . import _version

__version__ = _version.get_versions()["version"]
